/* V4L2 video picture grabber
   Copyright (C) 2009 Alan Carvalho de Assis
   Based on v4l2grab from Mauro Carvalho Chehab <mchehab@infradead.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   To compile:
   $ gcc gtkwc.c -o gtkwc `pkg-config --libs --cflags gtk+-2.0` -lv4l2

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <unistd.h>
#include <linux/videodev2.h>
#include "libv4l2.h"

#include <gtk/gtk.h>

#define CLEAR(x) memset(&(x), 0, sizeof(x))

#define IMAGE_WIDTH	329
#define IMAGE_HEIGHT	240
#define IMAGE_FILE "/usr/share/pixmaps/gtkwebcam/retrovisor.png"

int	fd = -1, frame = 0, n_buffers;
struct	v4l2_buffer	buf;
struct	buffer		*buffers;

guchar *rgbbuf;

void *capture_image_thread(void *ptr);

gint repaint (gpointer data);

gboolean on_darea_expose (GtkWidget *widget,
			  GdkEventExpose *event,
			  gpointer user_data);

static void destroy( GtkWidget *widget,
                     gpointer   data );

void draw_retrovisor(GtkWidget *widget);

GdkPixbuf* pixbuf;

struct buffer {
	void   *start;
	size_t length;
};

static void xioctl(int fh, int request, void *arg)
{
	int r;

	do {
		r = v4l2_ioctl(fh, request, arg);
	} while (r == -1 && ((errno == EINTR) || (errno == EAGAIN)));

	if (r == -1) {
		fprintf(stderr, "error %d, %s\n", errno, strerror(errno));
		exit(EXIT_FAILURE);
	}
}

int main(int argc, char **argv)
{
	struct v4l2_format		fmt;
	struct v4l2_requestbuffers	req;
	enum v4l2_buf_type		type;
	pthread_t			thread1;
	unsigned int			i, ret;
	char				*dev_name = "/dev/video0";

	GtkWidget *window, *darea;
        GError* error = NULL; 

	gtk_init (&argc, &argv);

	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	darea = gtk_drawing_area_new ();
	gtk_widget_set_size_request (darea, 800, 480);
	gtk_container_add (GTK_CONTAINER (window), darea);
	gtk_signal_connect (GTK_OBJECT (darea), "expose-event",
		GTK_SIGNAL_FUNC (on_darea_expose), NULL);

	g_signal_connect (G_OBJECT (window), "destroy",
		G_CALLBACK (destroy), NULL);

	gtk_widget_show_all (window);

	fd = v4l2_open(dev_name, O_RDWR | O_NONBLOCK, 0);
	if (fd < 0) {
		perror("Cannot open device");
		exit(EXIT_FAILURE);
	}

	CLEAR(fmt);
	fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fmt.fmt.pix.width       = IMAGE_WIDTH;
	fmt.fmt.pix.height      = IMAGE_HEIGHT;
	fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB24;
	fmt.fmt.pix.field       = V4L2_FIELD_INTERLACED;
	xioctl(fd, VIDIOC_S_FMT, &fmt);
	if (fmt.fmt.pix.pixelformat != V4L2_PIX_FMT_RGB24) {
		printf("Libv4l didn't accept RGB24 format. Can't proceed.\n");
		exit(EXIT_FAILURE);
	}
	if ((fmt.fmt.pix.width != IMAGE_WIDTH) || (fmt.fmt.pix.height != IMAGE_HEIGHT))
		printf("Warning: driver is sending image at %dx%d\n",
			fmt.fmt.pix.width, fmt.fmt.pix.height);

	CLEAR(req);
	req.count = 2;
	req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	req.memory = V4L2_MEMORY_MMAP;
	xioctl(fd, VIDIOC_REQBUFS, &req);

	buffers = calloc(req.count, sizeof(*buffers));
	for (n_buffers = 0; n_buffers < req.count; ++n_buffers) {
		CLEAR(buf);

		buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory      = V4L2_MEMORY_MMAP;
		buf.index       = n_buffers;

		xioctl(fd, VIDIOC_QUERYBUF, &buf);

		buffers[n_buffers].length = buf.length;
		buffers[n_buffers].start = v4l2_mmap(NULL, buf.length,
			      PROT_READ | PROT_WRITE, MAP_SHARED,
			      fd, buf.m.offset);

		if (MAP_FAILED == buffers[n_buffers].start) {
			perror("mmap");
			exit(EXIT_FAILURE);
		}
	}

	for (i = 0; i < n_buffers; ++i) {
		CLEAR(buf);
		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_MMAP;
		buf.index = i;
		xioctl(fd, VIDIOC_QBUF, &buf);
	}
	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	xioctl(fd, VIDIOC_STREAMON, &type);

	ret = pthread_create( &thread1, NULL, capture_image_thread, NULL);

	gtk_timeout_add (10, repaint, (gpointer) darea);

        pixbuf = gdk_pixbuf_new_from_file(IMAGE_FILE, &error ); 

	draw_retrovisor(window);

	gtk_main ();

	return 0;
}

void *capture_image_thread(void *ptr)
{
	fd_set				fds;
	struct timeval			tv;
	int				r, i = 0;

	CLEAR(buf);
	buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory = V4L2_MEMORY_MMAP;

	while(1){

		if(frame == 0){

		i++;

		/* queue another frame */
		xioctl(fd, VIDIOC_QBUF, &buf);

		do {
			FD_ZERO(&fds);
			FD_SET(fd, &fds);

			/* Timeout. */
			tv.tv_sec = 2;
			tv.tv_usec = 0;
			
			r = select(fd + 1, &fds, NULL, NULL, &tv);

		} while ((r == -1 && (errno = EINTR)));

		if (r == 0){
			printf("Count = %d\n", i);
			i = 0;
		}

		if (r == -1) {
			perror("select");
			exit(1);
		}

		/*if (r == 0){
			printf("Count = %d\n", i);
			i = 0;
		}*/

		/* dequeue the frame */
		xioctl(fd, VIDIOC_DQBUF, &buf);

		/* repaint() will draw new frame pointed by rgbbuf */
		rgbbuf = buffers[buf.index].start;

		frame = 1;

		}
		else{
			usleep(5000);
		}
	}
}

gint repaint (gpointer data) {
 
  GtkWidget* widget = (GtkWidget *) data;

  if(frame == 1){
   gdk_draw_rgb_image (widget->window, widget->style->fg_gc[GTK_STATE_NORMAL],
    240, 117, IMAGE_WIDTH, IMAGE_HEIGHT,
    GDK_RGB_DITHER_MAX, rgbbuf, IMAGE_WIDTH * 3);
    frame = 0;
  }

  return TRUE;

}


gboolean
on_darea_expose (GtkWidget *widget,
		 GdkEventExpose *event,
		 gpointer user_data)
{

  draw_retrovisor(widget);

  if(frame == 1){
    gdk_draw_rgb_image (widget->window, widget->style->fg_gc[GTK_STATE_NORMAL],
    240, 117, IMAGE_WIDTH, IMAGE_HEIGHT,
    GDK_RGB_DITHER_MAX, rgbbuf, IMAGE_WIDTH * 3);
    frame = 0;
  }

  return TRUE;
}

static void destroy( GtkWidget *widget,
                     gpointer   data )
{
	enum v4l2_buf_type	type;
	int i;

	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	xioctl(fd, VIDIOC_STREAMOFF, &type);
	for (i = 0; i < n_buffers; ++i)
		v4l2_munmap(buffers[i].start, buffers[i].length);
	v4l2_close(fd);
	gtk_main_quit ();
}

void draw_retrovisor(GtkWidget *widget)
{
	gdk_draw_pixbuf( widget->window, widget->style->fg_gc[GTK_STATE_NORMAL],
	pixbuf,	0, 0, 0, 0, -1, -1, GDK_RGB_DITHER_MAX, 0, 0);
	
}

